@@ -1,11 +1,8 @@
 """Various utilities used by the library"""
-# Standard library imports
 import time
 import warnings
-from typing import List
-from typing import Union
+from typing import List, Union
 
-# Third-party imports
 import numpy as np
 import scipy.interpolate as interpolate
 
